/*
 * GccApplication1.c
 *
 * Created: 10/24/2013 11:05:49 PM
 *  Author: Rolf R Bakke
 */ 




#define FOSC 16000000 // Clock Speed
#define BAUD 9600


#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdbool.h>
#include <stdlib.h>
#include <math.h>
#include "BitMacros.h"
#include "portDefs.h"
#include "servocal.h"

volatile uint8_t servoValues[6] = {s0_inactive, s1_inactive, s2_inactive, s3_inactive, s4_inactive, s5_inactive};
uint8_t charBuffer[32];
float colorRed, colorBlue, colorGreen, colorAll;
float colorHue, colorSpan;
uint8_t stations[5] = {0,0,0,0,0};
uint8_t stationsInactive[4] = {s2_inactive, s3_inactive, s4_inactive, s5_inactive};
uint8_t stationsActive[4] = {s2_active, s3_active, s4_active, s5_active};
	
void delayMicroseconds( unsigned long delay) {
	static unsigned long timeStart;
	unsigned long timeCurrent;
	
	
	if (delay == 0) {
		TCCR1B = bit(CS11) | bit(CS10);			//set TCNT1 to run at 250kHz.
		cli();
		timeStart = (unsigned long)TCNT1 << 2;
		sei();
		return;
	}
	
	//debugtimeUsedInDelayLoopOn;
	do {
		cli();
		timeCurrent = (unsigned long)TCNT1 << 2;
		sei();
	} while (((timeCurrent - timeStart) & (unsigned long)0x0003ffff) < delay);
	
	timeStart = timeCurrent;
	
	//debugtimeUsedInDelayLoopOff;
	
	return;
}

void error(int number) {
	delayMicroseconds(0);
	
	while(1){
		for (int i = 0; i < number; i++) {
			ledOn;
			delayMicroseconds(250000);
			ledOff;
			delayMicroseconds(250000);
		}
		delayMicroseconds(250000);
		delayMicroseconds(250000);
		delayMicroseconds(250000);
		delayMicroseconds(250000);
	}
}

void USART_Init( unsigned int ubrr){
	/* Set baud rate */
	UBRR0 = ubrr;
	/* Enable transmitter */
	UCSR0B = (1<<TXEN0);
	/* Set frame format: 8data, 1stop bit */
	UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);
}

void serialSendChar(unsigned char data) {
	while ((UCSR0A & 1<<UDRE0) == 0);			//wait until TX data is empty.
	UDR0 = data;
}

void serialSendString(char *input) {
	char data;
	while ((data = *input++)) serialSendChar(data); 
}
	
ISR(TIMER1_COMPA_vect) {
	//ledOn;
	static uint8_t state = 0;
	static uint16_t currentTime = 0;
	static uint16_t currentFrame;
	uint8_t bitMask;
	uint8_t i;
	
	if (state == 0) currentFrame = currentTime;  //we are at start of a frame. Save current time for calculating start of next frame
	
	if (state == 6) {		//we are at the end of current frame
		state = 0;			//set state to start of frame (servo #0)
		PORTC = 0x00;		//turn off last servo output (servo #5)
		currentTime = currentFrame + 5000;  //set time for start of next frame.
		OCR1A = currentTime;
		//ledOff;
		return;
	}
	
	bitMask = 0b00000001;
	i = state;				//get servo to be updated;
	while (i > 0) {			//create bit mask for that servo
		bitMask = bitMask<<1;
		i--;
	}
	PORTC = bitMask;		//turn on output for that servo
	currentTime = currentTime + 250 + servoValues[state];	//set time for next servo/pulse duration for this servo
	OCR1A = currentTime;
	state++;				//advance to next servo
	//ledOff;
	return;
}	
	
	
void updateLCD(void) {
	serialSendString(" $GO 1 1\n $PRINT ");
	for (uint8_t i = 0; i<32; i++) {
		if (i == 16) serialSendString("\n $GO 2 1\n $PRINT ");
		serialSendChar(charBuffer[i]);
	}
	serialSendString("\n");
}

void clearLCD(void) {
	for (uint8_t i = 0; i<32; i++) {
		charBuffer[i] = ' ';
	}
}

void printLCD( uint8_t position, char *input) {
	char data;
	uint8_t i = 0;
	while ((data = *input++)) charBuffer[(i++) + position] = data;
}

void printNumberLCD(uint8_t position, uint16_t number) {
	bool firstNonZero = false;
	uint16_t dividers[5] = {10000, 1000, 100, 10, 1};
		
	for (uint8_t i = 0; i<=4; i++) {
		uint8_t digit = number / dividers[i];
		number -= digit * dividers[i];
		if (digit) firstNonZero = true;
		if (firstNonZero || i == 4) charBuffer[position++] = digit + 0x30;
	}	
}

bool inRange(uint8_t value, uint8_t target, uint8_t offset){
	if ((value > (target - offset)) && (value < (target + offset))) return true;
	return false;
}

uint8_t getColor(void){
	float value;
	float min = 10000;
	float max = 0;
	float gain;

	for (uint8_t i = 0; i <= 2; i++){
		if (i==0) PORTD = (PORTD & 0b11010011) | 0b00000100;	//red
		if (i==1) PORTD = (PORTD & 0b11010011) | 0b00001000;	//green
		if (i==2) PORTD = (PORTD & 0b11010011) | 0b00100000;	//blue
		
		delayMicroseconds(0);
		delayMicroseconds(1000);
		
		TCNT0 = 0x00;				//clear counter
		TIFR0 =  0b00000001;		//clear overflow flag
		//         76543210
		TCCR0B = 0b00000111;		//start counting

		delayMicroseconds(0);		//wait x ms (any multiple of the 100Hz light freq)
		delayMicroseconds(250000);
		delayMicroseconds(60000);

		//         76543210
		TCCR0B = 0b00000000;		//stop counting
	
		value = TCNT0;
		if (TIFR0 & 0b00000001)	value = 0xff; //limit to 255 if overflow
		
		if (i==0) colorRed = value * 2.103;
		if (i==1) colorGreen = value * 1.538;
		if (i==2) colorBlue = value * 1;		
	}	
	
	PORTD = (PORTD & 0b11010011);	//turn of all colors

	colorAll = colorRed + colorGreen + colorBlue;

	if (colorRed < min) min = colorRed;
	if (colorRed > max) max = colorRed;
	if (colorBlue < min) min = colorBlue;
	if (colorBlue > max) max = colorBlue;
	if (colorGreen < min) min = colorGreen;
	if (colorGreen > max) max = colorGreen;
		
	gain = 100.0 / max;				//normalize to 100 
	//gain = 1;
	colorRed = colorRed * gain;
	colorGreen = colorGreen * gain;
	colorBlue = colorBlue * gain;

	colorSpan = (max - min) * gain;
	//colorHue = (atan2(1.73205 * (colorGreen - colorBlue), 2.0 * colorRed - colorGreen - colorBlue) + 3.1415) * 57.296;

	clearLCD();
	printNumberLCD(0, colorRed);
	printNumberLCD(4, colorGreen);
	printNumberLCD(8, colorBlue);
	printNumberLCD(12, colorAll);
	printNumberLCD(28, colorSpan);

	if ((colorSpan < 50) && (colorAll > 150)) {		//nothing	
		printLCD(16, "Nothing");
		return 0;
	}	
	if ((colorSpan < 50) && (colorAll < 150)) {		//black	
		printLCD(16, "Black");
		return 5;
	}
	if (inRange(colorRed, 100, 10) && inRange(colorGreen, 60, 10) && inRange(colorBlue, 17, 10)){
		printLCD(16, "Yellow");
		return 1;
	}
	if (inRange(colorRed, 42, 10) && inRange(colorGreen, 100, 10) && inRange(colorBlue, 31, 10)){
		printLCD(16, "Green");
		return 2;
	}
	if (inRange(colorRed, 100, 10) && inRange(colorGreen, 15, 10) && inRange(colorBlue, 17, 10) && (colorAll < 172)){
		printLCD(16, "Red");
		return 3;
	}
	if (inRange(colorRed, 100, 10) && inRange(colorGreen, 26, 10) && inRange(colorBlue, 18, 10) && (colorAll >= 172)){
		printLCD(16, "Orange");
		return 4;
	}

	printLCD(16, "Unknown");
	return 0;
}

void delayMilliseconds(int16_t value){
	delayMicroseconds(0);
	while (value > 0){
		delayMicroseconds(1000);
		value--;
	}
}




int main(void)
{

	//       76543210	
	DDRD = 0b00101100;
	DDRC = 0b00111111;
	DDRB = 0b00100000;



	//HW init for the servo driver
	//         76543210
	TCCR1B = 0b00000011; //start timer
	TIMSK1 = 0b00000010; //enable OCR1a
	sei();

	USART_Init (FOSC/16/BAUD-1 );

	//goto calibration;

	while(1){
		
		servoValues[0] = s0_pos2;			//feed one nonstop
		delayMilliseconds(800);
		servoValues[0] = s0_inactive;
		
		servoValues[1] = s1_active;			//move belt one position
		delayMilliseconds(800);
		servoValues[1] = s1_inactive;	
		
		for (int i = 4; i >= 1; i--){		//move station array one position
			stations[i] = stations[i-1];
		}
		
		for (int i = 0; i <= 3; i++){		//kick off nonstops that are at correct station
			if (stations[i+1] == i+1) servoValues[i+2] = stationsActive[i];
		}
		delayMilliseconds(800);
		for (int i = 0; i <= 3; i++){
			servoValues[i+2] = stationsInactive[i];
		}
		
		stations[0] = getColor();			//get color at station 0;
	
		updateLCD();
		
		
		//while(1);
	}




	
	

calibration:	
	while(1) {
	getColor();
	clearLCD();	
	printNumberLCD(0, colorRed);
	printNumberLCD(4, colorGreen);
	printNumberLCD(8, colorBlue);
	printNumberLCD(12, colorAll);
	printNumberLCD(16, colorHue);
	printNumberLCD(20, colorSpan);
	updateLCD();
	}	
}